/*
 * (c) Copyright 2001-2003
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Version: 2.1.0.0 - Aug 25 2003 11:11:11
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "s68.h"
#include "device.h"
#include "decoder.h"
#include "conf.h"


#define REGSIZE     32


#define DATA          0x01       /* Data register */
#define ERROR         0x03       /* Error register */
#define SECCNT        0x05       /* Sector count register */
#define SECNUM        0x07       /* Secotr number register */
#define CYLLOW        0x09       /* Cylinder Low register */
#define CYLHIGH       0x0B       /* Cylinder High register */
#define SDH           0x0D       /* Select Drive and Head Number */
#define CMD           0x0F       /* Command register (write only) */
#define STAT          0x0F       /* Status (read only) */

#define CMD_READ      0x20       /* Read Sector */
#define CMD_WRITE     0x30       /* Write Sector */
#define CMD_FMTRACK   0x50       /* Format Track */
#define CMD_RDVFY     0x40       /* Read Verify */
#define CMD_EXDIAG    0x90       /* Execute Diagnostic */
#define CMD_SETPARA   0x91       /* Set Drive Parameter */
#define CMD_READM     0xC4       /* Read Multiple Sectors */
#define CMD_WRITEM    0xC5       /* Write Multiple Sectors */
#define CMD_SETM      0xC6       /* Set Multiple */
#define CMD_STBYIM    0xE0       /* Standby Immediate */
#define CMD_IDLEIM    0xE1       /* Idle Immediate */
#define CMD_STANDBY   0xE2       /* Standby */
#define CMD_IDLE      0xE3       /* Idle */
#define CMD_READBUF   0xE4       /* Read Buffer */
#define CMD_CHKPMODE  0xE5       /* Check Power Mode */
#define CMD_SLEEP     0xE6       /* Sleep */
#define CMD_REST      0xE7       /* Rest */
#define CMD_WRTBUF    0xE8       /* Write Buffer */
#define CMD_RDSTATE   0xE9       /* Read Drive State */
#define CMD_RSSTATE   0xEA       /* Restore Drive State */
#define CMD_IDENTIFY  0xEC       /* Identify Drive */
#define CMD_SETFEAT   0xEF       /* Set Features */

#define STAT_ERR      0x01       /* Unrecoverable error */
#define STAT_IDX      0x02       /* Index pulse of target drive */
#define STAT_CORR     0x04       /* Corrected Data */
#define STAT_DRQ      0x08       /* Data Request */
#define STAT_DSC      0x10       /* Drive Seek Complete */
#define STAT_DWF      0x20       /* Drive Write Fault */
#define STAT_DRDY     0x40       /* Ready state of target drive */
#define STAT_BUSY     0x80       /* Busy state of controller */

#define ERR_AMNF      0x01       /* Address Mark Not Found */
#define ERR_TK0NF     0x02       /* Track 0 Not Found */
#define ERR_ABRT      0x04       /* Aborted Command */
#define ERR_MCR       0x08       /* Media Change Requested */
#define ERR_IDNF      0x10       /* ID Not Found */
#define ERR_MC        0x20       /* Media Changed */
#define ERR_UNC       0x40       /* Uncorrectable Data Error */
#define ERR_BBD       0x80       /* Bad Block Detected */

#define HEADS         8          /* Number of heads */
#define SECPTRK       32         /* Sectors per Track */


typedef struct  {
    u_char  data;
    u_char  error;
    u_char  seccnt;
    u_char  secNum;
    u_char  cylLow;
    u_char  cylHigh;
    u_char  sdh;
    u_char  cmd;
    u_char  stat;

    u_char  sectorBuffer[1024];
    u_long  sectorIdx;

    u_short driveParam[512/2];
    u_long  diskSize;

    int  numOfHeads;
    int  numOfSecpTrk;
    int  numOfCylinder;

    FILE *imgFile;
} ATA;

 

/*
 * s w a p B u f f e r
 *
 * Swap bytes of buffer
 */
static void swapBuffer(ATA *ata)
{
    u_short  *wp = (unsigned short *) ata->sectorBuffer;
    int   i;

    for (i = 0; i < 512/2; i++)  {
         *wp = ((*wp << 8) | (*wp >> 8));
         wp++;
    }
}



/*
 * t r a n s f e r S e c t o r
 *
 * Transfer sector buffer to/from disk
 */
static u_char transferSector(ATA *ata, u_char mode)
{
    u_long  lba;
    u_long  fileOffset;
    off_t   ret;
    size_t  cnt;


    lba = (((ata->cylHigh << 8) + ata->cylLow) * (ata->numOfHeads * ata->numOfSecpTrk)) \
         + ((ata->sdh & 0x0F) * ata->numOfSecpTrk) \
         + (ata->secNum -1);
    fileOffset = lba * 512;

    ret = fseek(ata->imgFile, fileOffset, SEEK_SET);
    if (ret == (off_t) -1)  {
        /* Seek error */
        return(ERR_IDNF);
    }

    switch (mode)  {
        case CMD_READ :
            cnt = fread(ata->sectorBuffer, 1, 512, ata->imgFile);
            if (cnt != (size_t) 512)  {
                /* Read error */
                uiErrPrintf("ATA: Read - %s\n", strerror(errno));
                return(ERR_UNC);
            }
            swapBuffer(ata);
            break;

        case CMD_WRITE :
            swapBuffer(ata);
            cnt = fwrite(ata->sectorBuffer, 1, 512, ata->imgFile);
            if (cnt != (size_t) 512)  {
                /* Write error */
                uiErrPrintf("ATA: Write - %s\n", strerror(errno));
                return(ERR_BBD);
            }
            break;

        default :
            return(ERR_ABRT);

    }

    return (0);
}



/*
 * f o r m a t T r a c k
 *
 * Format track
 */
static u_char formatTrack(ATA *ata)
{
    u_long  lba;
    u_long  fileOffset;
    off_t   ret;


    lba = (((ata->cylHigh << 8) + ata->cylLow) * (ata->numOfHeads * ata->numOfSecpTrk)) \
         + ((ata->sdh & 0x0F) * ata->numOfSecpTrk);
    fileOffset = lba * 512;

    ret = fseek(ata->imgFile, fileOffset, SEEK_SET);
    if (ret == (off_t) -1)  {
        /* Seek error */
        return(ERR_IDNF);
    }

    return (0);
}

 

/*
 * a t a R e s e t
 *
 * Reset the device
 */
int ataReset(Device *dev)
{
    ATA   *ata = (ATA *) dev->priv;

    if (!quiet)  uiPrintf(" ATA");

    ata->numOfHeads = HEADS;
    ata->numOfSecpTrk = SECPTRK;
    ata->numOfCylinder = ata->diskSize / (ata->numOfSecpTrk * ata->numOfHeads);

    /* Initialize drive Parameters */
    memset(ata->driveParam, 0x00, 512);
    ata->driveParam[0]  = 0x0044;              /* fixed drive, soft sectored */
    ata->driveParam[1]  = ata->numOfCylinder;  /* Number of Cylinder */
    ata->driveParam[3]  = HEADS;               /* Number of Heads */
    ata->driveParam[4]  = SECPTRK*520;         /* Number of unformatted bytes per track */
    ata->driveParam[5]  = 520;                 /* Number of unformated bytes per sector */
    ata->driveParam[6]  = SECPTRK;             /* Number of Secotors per Track */
    ata->driveParam[22] = 0x0008;              /* Number of ECC bytes in long commands */
    ata->driveParam[47] = 0x0008;              /* Maximum of 8 sectors in read/write mult. */
    ata->driveParam[49] = 0x0200;              /* LBA supported, DMA not supported */
    ata->driveParam[54] = ata->numOfCylinder;  /* Number of current cylinder */
    ata->driveParam[55] = HEADS;               /* Number of current heads */
    ata->driveParam[56] = SECPTRK;             /* Number of current secotors per track */
    ata->driveParam[57] = (u_short)(ata->diskSize & 0xFFFF);
    ata->driveParam[58] = (u_short)(ata->diskSize >> 16);
    ata->driveParam[59] = 0x0108;              /* Current number of multible sectors */
    ata->driveParam[60] = (u_short)(ata->diskSize & 0xFFFF);
    ata->driveParam[61] = (u_short)(ata->diskSize >> 16);

    /* Reset registers */
    ata->error = 0x01;         /* Selftest complete */
    ata->cmd = ata->stat = 0x00;
    ata->seccnt = ata->secNum = 0x01;
    ata->cylLow = ata->cylHigh = ata->sdh = 0x00;

    return 0;
}



/*
 * a t a R e a d
 *
 * Read one byte
 */
static u_long ataRead(Device *dev, u_long offset)
{
    ATA    *ata = (ATA *) dev->priv;
    u_char data;

    switch (offset & (REGSIZE-1))  {
        case DATA :
            if ((ata->cmd == CMD_READ) || (ata->cmd == CMD_IDENTIFY)) {
                data = ata->sectorBuffer[ata->sectorIdx];
                if (++ata->sectorIdx >= 512) {
                    /* End of read command */
                    ata->cmd = 0;
                    ata->stat = 0;
                }
            }
            else if (ata->cmd == CMD_WRITE)  {
                data = 0x55;
            }
            else  {
                data = 0xAA;
            }
            break;

        case ERROR :
            data = ata->error;
            break;

        case STAT :
            data = ata->stat;
            break;

        case SECCNT :
            data = ata->seccnt;
            break;

        case SECNUM :
            data = ata->secNum;
            break;

        case CYLLOW :
            data = ata->cylLow;
            break;

        case CYLHIGH :
            data = ata->cylHigh;
            break;
 
        case SDH :
            data = ata->sdh;
            break;

        default :
            data = 0xFF;
            break;
    }

    return (u_long) data;
}



/*
 * a t a W r i t e
 *
 * Write one byte
 */
static void ataWrite(Device *dev, u_long offset, u_char data)
{
    ATA    *ata = (ATA *) dev->priv;

    switch (offset & (REGSIZE-1))  {
      case DATA :
        if (ata->cmd == CMD_WRITE)  {
            ata->sectorBuffer[ata->sectorIdx] = data;
            if (++ata->sectorIdx >= 512)  {
                /* Buffer filled. Transfer to disk */
                ata->stat = STAT_BUSY;
                if ((ata->error = transferSector(ata, CMD_WRITE)) != 0)  {
                    /* Error during transfer to disk */
                    ata->stat = (ata->error == ERR_IDNF) ? STAT_ERR : STAT_DWF;
                }
                else  {
                    /* No error */
                    ata->stat = 0x00;
                }
                /* End of write command */
                ata->cmd = 0;
            }
        }
        else if (ata->cmd == CMD_FMTRACK)  {
            ata->sectorBuffer[ata->sectorIdx] = data;
            if (++ata->sectorIdx >= 512)  {
                /* Buffer filled. Transfer to disk */
                ata->stat = STAT_BUSY;
                if ((ata->error = formatTrack(ata)) != 0)  {
                    /* Error during transfer to disk */
                    ata->stat = (ata->error == ERR_IDNF) ? STAT_ERR : STAT_DWF;
                }
                else  {
                    /* No error */
                    ata->stat = 0x00;
                }
                /* End of format track command */
                ata->cmd = 0;
            }
        }
         break;

      case ERROR :
        ata->error = data;
        break;

      case SECCNT :
        ata->seccnt = data;
        break;

      case SECNUM :
        ata->secNum = data;
        break;

      case CYLLOW :
        ata->cylLow = data;
        break;

      case CYLHIGH :
        ata->cylHigh = data;
        break;

      case SDH :
        ata->sdh = data;
        break;

      case CMD :
        switch (data)  {
          case CMD_READ :
            ata->cmd = CMD_READ;
            ata->stat = STAT_BUSY;
            if ((ata->error = transferSector(ata, CMD_READ)) != 0)  {
                 /* Error during read */
                 ata->stat = STAT_ERR;
                 ata->cmd = 0x00;
            }
            else  {
                 /* Sector buffer filled */
                 ata->sectorIdx = 0;
                 ata->stat = STAT_DRQ;
                 ata->error = 0;
            }
            break;

          case CMD_WRITE :
            ata->cmd = CMD_WRITE;
            ata->stat = STAT_DRQ;
            ata->error = 0;
            ata->sectorIdx = 0;
            break;

          case CMD_IDENTIFY :
            ata->cmd = CMD_IDENTIFY;
            ata->stat = STAT_BUSY;
            memcpy(ata->sectorBuffer, ata->driveParam, 512);
            ata->sectorIdx = 0;
            ata->stat = STAT_DRQ;
            ata->error = 0x00;
            break;

          case CMD_FMTRACK :
            ata->stat = STAT_DRQ;
            ata->error = 0;
            ata->sectorIdx = 0;
            break;

          default :
            uiPrintf("ATA: command 0x%02x not implemented\n", data);
            ata->cmd = 0x00;
            ata->stat = STAT_ERR;
            ata->error = ERR_ABRT;
            break;
        }
        break;
    }

}



/*
 * i n i t A T A
 *
 * Initialize
 */
int initATA(Device *dev, u_long base, u_long size, void *arg)
{
    ATA   *ata;
    char  *fileName = (char *)arg;
    FILE  *imgFile;
    struct stat statBuffer;

    if ((ata = malloc(sizeof(ATA))) == NULL)  {
        uiErrPrintf("ATA: Malloc of %ld bytes failed - %s\n", size, strerror(errno));
        return(errno);
    }

    /* Open image file */
    if ((imgFile = fopen(fileName, "r+b")) == NULL)  {
        uiErrPrintf("ATA: Can't open %s - %s\n", fileName, strerror(errno));
        return(errno);
    }

    if (stat(fileName, &statBuffer) == -1)  {
        uiErrPrintf("ATA: Can't get size of %s - %s\n", fileName, strerror(errno));

        return(errno);
    }

    ata->imgFile  = imgFile;
    ata->diskSize = statBuffer.st_size / 512;

    /* Setup device structure */
    dev->base = (u_long) base;
    dev->mask = ~(size-1);
    dev->priv = (u_char *) ata;
    dev->devReset = ataReset;
    dev->devRead  = ataRead;
    dev->devWrite = ataWrite;
    dev->devIAck  = NULL;
    dev->devCycle = NULL;

    return 0;
}
